! Copyright (c) 2017, Tsutomu Kato, SofTek Systems. Inc. All rights reserved.
!
! Permission is hereby granted, free of charge, to any person obtaining a
! copy of this software and associated documentation files (the "Software"),
! to deal in the Software without restriction, including without limitation
! the rights to use, copy, modify, merge, publish, distribute, sublicense,
! and/or sell copies of the Software, and to permit persons to whom the
! Software is furnished to do so, subject to the following conditions:
!
! The above copyright notice and this permission notice shall be included in
! all copies or substantial portions of the Software.
!
! THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
! IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
! FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
! THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
! LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
! FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
! DEALINGS IN THE SOFTWARE.
!

module Matrix_data
! COO storage data and CSR data -- declare variables
  use precision
  implicit none

! Buffer for Matrix Market data format (as we say "COO" format)
    integer      :: rows, cols, nnz, nnzmax
    character*10 :: rep
    character*7  :: field
    character*19 :: symm
    integer, allocatable, dimension(:)       :: indx, jndx
    integer, allocatable, dimension(:)       :: ival
    real(fp_kind), allocatable, dimension(:) :: rval
    complex*16, allocatable, dimension(:)    :: cval

! Arrays for CUDA cuSPARSE CSR format (available for REAL data only)
    integer  :: m                            ! rows size of A matrix
    integer  :: n                            ! columns size of A matrix
    integer  :: nz                           ! nonzero element of A
    integer, allocatable, dimension(:)       :: csrRowPtrA  ! m + 1
    integer, allocatable, dimension(:)       :: csrColIndA  !
    real(fp_kind), allocatable, dimension(:) :: csrValA
!
    real(fp_kind), allocatable, dimension(:) :: x   ! initial guess & result vector
    real(fp_kind), allocatable, dimension(:) :: rhs ! right hand side vector

end module Matrix_data

module mm_coo_cuda_csr_convert
  use precision
  use Matrix_data
  use mmio
  implicit none
  
  contains

  subroutine matrix_printout(f, nonzero, i_ndx, j_ndx, rval1, ival1, cval1)
  implicit none
      integer                    :: i, nonzero
      character*7				 :: f
      integer                    :: i_ndx(nonzero)
      integer                    :: j_ndx(nonzero)
      integer,          optional :: ival1(nonzero)
      double precision, optional :: rval1(nonzero)
      complex*16,       optional :: cval1(nonzero)

      do i = 1, min(nonzero,10)
        if (f == "complex") then
          print *,"i,indx,jndx",i, i_ndx(i),j_ndx(i),cval1(i)
        elseif (f == "real") then
          print *,"i,indx,jndx",i, i_ndx(i),j_ndx(i),rval1(i)
        elseif (f == "integer") then
          print *,"i,indx,jndx",i, i_ndx(i),j_ndx(i),ival1(i)
        end if
      end do
      do i = max(nonzero-10, 1), nonzero
        if (f == "complex") then
          print *,"i,indx,jndx",i, i_ndx(i),j_ndx(i),cval1(i)
        elseif (f == "real") then
          print *,"i,indx,jndx",i, i_ndx(i),j_ndx(i),rval1(i)
        elseif (f == "integer") then
          print *,"i,indx,jndx",i, i_ndx(i),j_ndx(i),ival1(i)
        end if
      end do
      
  end subroutine matrix_printout

  subroutine csr_printout
  implicit none
    integer :: i
      print *
      print *,"== CSR data check : Start from 1  to 10 elements =="
      if (field == "complex") then
        print '(1x, a/, 7(1x,I10   ))',"csrRowPtrA=",(csrRowPtrA(i), i=1,min(m+1,10))
        print '(1x, a/, 7(1x,I10   ))',"csrColIndA=",(csrColIndA(i), i=1,min(nz,10))
        print '(1x  a/, 7(1x,2d10.4))',"csrValA  =",(csrValA(i), i=1,min(nz,10))
      elseif (field == "real") then
        print '(1x, a/, 7(1x,I10  ))',"csrRowPtrA=",(csrRowPtrA(i), i=1,min(m+1,10))
        print '(1x, a/, 7(1x,I10  ))',"csrColIndA=",(csrColIndA(i), i=1,min(nz,10))
        print '(1x, a/, 7(1x,d10.4))',"csrValA  =",(csrValA(i), i=1,min(nz,10))
      end if

      print *
      print *,"== Start from endpoint-10  to endpoint =="
      if (field == "complex") then
        print '(1x, a/, 7(1x,I10   ))',"csrRowPtrA=",(csrRowPtrA(i), i=max(m+1-10,1),m+1)
        print '(1x, a/, 7(1x,I10   ))',"csrColIndA=",(csrColIndA(i), i=max(nz-10,1),nz)
        print '(1x, a/, 7(1x,2d10.4))',"csrValA=",(csrValA(i), i=max(nz-10,1),nz)
      elseif (field == "real") then
        print '(1x, a/, 7(1x,I10  ))',"csrRowPtrA=",(csrRowPtrA(i), i=max(m+1-10,1), m+1)
        print '(1x, a/, 7(1x,I10  ))',"csrColIndA=",(csrColIndA(i), i=max(nz-10,1),nz)
        print '(1x, a/, 7(1x,d10.4))',"csrValA=",(csrValA(i), i=max(nz-10,1),nz)
      end if

  end subroutine csr_printout

  subroutine set_matrix(iunit1, extendSymMatrix, field, symm, t_nnz, t_indx, t_jndx, &
                        t_ival, t_rval, t_cval)
  use precision
  use Matrix_data
 
    character*7             :: field
    character*19            :: symm
    integer                 :: t_nnz
    integer ,optional       :: t_indx(t_nnz), t_jndx(t_nnz)
    integer, optional       :: t_ival(t_nnz)
    real(fp_kind), optional :: t_rval(t_nnz)
    complex*16, optional    :: t_cval(t_nnz)

    integer                 :: i, j, iunit1, count
    integer                 :: extendSymMatrix   ! 1 = extend , 0 = no

    
        if (symm == "symmetric" .and. extendSymMatrix == 1) then
          ! count number of non-diagonal elements
          count = 0
          do i = 1, t_nnz
            if (t_indx(i) /= t_jndx(i)) count=count+1
          end do
          allocate(indx(count + t_nnz))
          allocate(jndx(count + t_nnz))
            if (field == "real"   ) allocate(rval(count + t_nnz))
            if (field == "integer") allocate(ival(count + t_nnz))
            if (field == "complex") allocate(cval(count + t_nnz))
          ! copy the elements regular and transposed locations
          j = 1 
          do i = 1, t_nnz
            indx(j) = t_indx(i)
            jndx(j) = t_jndx(i)
            if (field == "real"   ) rval(j) = t_rval(i)
            if (field == "integer") ival(j) = t_ival(i)
            if (field == "complex") cval(j) = t_cval(i)
            j = j + 1
            if (t_indx(i) /= t_jndx(i)) then
              indx(j) = t_jndx(i)
              jndx(j) = t_indx(i)
              if (field == "real"   ) rval(j) = t_rval(i)
              if (field == "integer") ival(j) = t_ival(i)
              if (field == "complex") cval(j) = t_cval(i)
              j = j + 1
            end if
          end do
          t_nnz = t_nnz + count
        else
          allocate(indx(t_nnz))
          allocate(jndx(t_nnz))
          allocate(cval(t_nnz))
          indx(:) = t_indx(:)
          jndx(:) = t_jndx(:)
          if (field == "real"   ) then 
            allocate(rval(t_nnz))
            rval(:) = t_rval(:)
          end if 
          if (field == "integer") then
            allocate(ival(t_nnz))
            ival(:) = t_ival(:)
          end if
          if (field == "complex") then
            allocate(cval(t_nnz))
            cval(:) = t_cval(:)
          end if
        end if

  end subroutine set_matrix
  
  subroutine mm_matrix_read(iunit1, extendSymMatrix)

!   Read Matrix Market format file

    implicit none
    integer, allocatable, dimension(:)       :: t_indx, t_jndx
    integer, allocatable, dimension(:)       :: t_ival
    real(fp_kind), allocatable, dimension(:) :: t_rval
    complex*16, allocatable, dimension(:)    :: t_cval
    integer      :: i, iunit1, count
    integer      :: extendSymMatrix   ! 1 = extend , 0 = none

!   Read header infos
      call mminfo(iunit1, rep, field, symm, rows, cols, nnz)

       ! small test data !
       ! !       | 1 2 0 |
       ! !   A = | 0 5 0 |   base = 1
       ! !       | 0 8 0 |
       ! rep   = "coordinate"
       ! field = "real"
       ! symm  ="general"
       ! rows = 3 ; cols = 3 ; nnz = 4
       ! indx(1:4) = (/3, 2, 1, 1/) 
       ! jndx(1:4) = (/2, 2, 1, 2/) 
       ! rval(1:4) =(/8.0, 5.0, 1.0, 2.0/)  ! for small test data

      nnzmax = nnz
      allocate (t_indx(nnz), t_jndx(nnz))

      if (field == "complex") then
        allocate (t_cval(nnz))
        call mmread(iunit1, rep, field, symm, rows, cols, nnz, nnzmax, &
                    t_indx, t_jndx, cval=t_cval)
        call set_matrix(iunit1, extendSymMatrix, field, symm, nnz, t_indx, t_jndx, &
                    t_cval=t_cval)
      elseif (field == "real") then
        allocate (t_rval(nnz))
        call mmread(iunit1, rep, field, symm, rows, cols, nnz, nnzmax, &
                    t_indx, t_jndx, rval=t_rval)
        call set_matrix(iunit1, extendSymMatrix, field, symm, nnz, t_indx, t_jndx, &
                    t_rval=t_rval)
      elseif (field == "integer") then
        allocate (t_ival(nnz))
        call mmread(iunit1, rep, field, symm, rows, cols, nnz, nnzmax, &
                    t_indx, t_jndx, ival=t_ival)
        call set_matrix(iunit1, extendSymMatrix, field, symm, nnz, t_indx, t_jndx, &
                    t_ival=t_ival)
      elseif (field == "pattern") then
        print *, "Pattern field is not available." ; stop "10"
      else
        print *, "Error : data type not found." , field ; stop "10"
      end if

        deallocate(t_indx,t_jndx)
        if (field == "real"   ) deallocate(t_rval)
        if (field == "integer") deallocate(t_ival)
        if (field == "complex") deallocate(t_cval)

      print *,"Properties of Matrix_Market matrix"
      print *,"===================================="
      print '(3(1x,a,1x,a/),4(1x,a,i8/))', &
               "representation =",rep,  &
               "type           =",field,&
               "matrix type    =",symm, &
               "row (m)        =",rows, &
               "cols(n)        =",cols, &
               "Non-zero elm.  =",nnz,  &
               "extendSymMatrix=",extendSymMatrix
      if (symm == "symmetric" .and. extendSymMatrix == 1) &
        print '(/a/)', " ==== Triangular Symmetric matrix has been transformed to full matrix form ===="
      

      if (field == "complex") then
        call matrix_printout (field, nnz, indx, jndx, cval1=cval)
      elseif (field == "real") then
        call matrix_printout (field, nnz, indx, jndx, rval1=rval)
      elseif (field == "integer") then
        call matrix_printout (field, nnz, indx, jndx, ival1=ival)
        print *, "No available in case of field = integer in this routine"
        stop "20"
      elseif (field == "pattern") then
        print *, "No available in case of field = pattern in this routine"
        stop "30"
      end if
      close(iunit1)

  end subroutine mm_matrix_read

  subroutine mm_rhs_read(iunit2)

!   INPUT   : vector file (Matrix Market format)
!   OUTPUT  : rhs(:) -- global array

    use mmio
    implicit none
    integer      :: i, iunit2
    character*40 :: rhs_file
    integer      :: rows1, cols1, nnz1, nnzmax1
    character*10 :: rep1
    character*7  :: field1
    character*19 :: symm1
    integer, allocatable, dimension(:) :: indx1, jndx1

      call mminfo(iunit2, rep1, field1, symm1, rows1, cols1, nnz1)
      
      if (rows1 /= rows) then
         print *, "Error : RHS vector size does match with Matrix row size."
         stop
      end if

      print *,"Properties of Matrix_Market RHS vector"
      print *,"===================================="
      print '(3(1x,a,1x,a/),3(1x,a,i8/))', &
               "representation =",rep1,  &
               "type           =",field1,&
               "matrix type    =",symm1, &
               "row (m)        =",rows1, &
               "cols(n)        =",cols1, &
               "Non-zero elm.  =",nnz1
      
      nnzmax1 = nnz1
      allocate (indx1(rows1), jndx1(cols1))
      allocate (rhs(nnz1))
      call mmread(iunit2, rep1, field1, symm1, rows1, cols1, nnz1, nnzmax1, &
                  indx1, jndx1, rval=rhs)
      ! print *,"rhs=", rhs
      deallocate (indx1, jndx1)
      close(iunit2)

  end subroutine mm_rhs_read

  subroutine coo2csr_convert

!   INPUT   : Matrix COO data from Matrix Market format via mm_read routine
!   OUTPUT  : CSR data = csrRowPtrA(m+1), csrColIndA(nz), csrValA(nz) -- global data

    use cublas
    use cusparse
                                         !! fortran program as index_base = 1

    type(cusparseHandle)                 :: cusparse_H
    type(cublasHandle)                   :: cublas_H
    integer(8)                           :: buff_size  ! must be 8byte integer. so needs -Mlarge_arrays option
    character*1,allocatable,dimension(:) :: pBuffer
    integer,    allocatable,dimension(:) :: P 
    integer                              :: status
    integer                              :: ierr_code
    integer                              :: i

      if ( rows /= cols ) then
        print *, "Error : Matrix must be square (rows = cols) in this routine."
        stop
      end if
     
    ! set m, n, nz
      m = rows    
      n = cols    ! matrix size 
      nz = nnz    ! nz == nnz
      if ( m /= n ) then
        print *, "Error : Matrix A is not square matrix ",m,"x",n
        stop 
      end if

    ! set CSR storage arrays
      allocate(csrRowPtrA(m+1))
      allocate(csrColIndA(nz) )
      allocate(csrValA(nz)    )
    ! Initilize
      csrRowPtrA =0 ; csrColIndA =0 ; csrValA = 0_fp_kind
      
    ! index base is changed from 1 to 0 for using cusparseXcoosortByRow routine
      do i = 1, nz
          indx(i) = indx(i) - 1
          jndx(i) = jndx(i) - 1
      end do

        !! call matrix_printout (field, nnz, indx, jndx, rval1=rval)

    ! create handle for CUSPARSE
      status = cusparseCreate(cusparse_H)

    !$acc data copy(indx, jndx) copy(rval, csrValA, csrRowPtrA)

      ! calculate buffer size to conver
        !$acc host_data use_device(indx, jndx)
          status = cusparseXcoosort_bufferSizeExt(cusparse_H, m, n, nz, indx, jndx, buff_size)
        !$acc end host_data
        print *, "Xcoosort_bufferSizeExt status =", status 

      ! Prepare working space
        print *, "working space (bytes) =", buff_size
        allocate(pBuffer(buff_size), stat=ierr_code)
        print *, "allocation status code for pBuffer =", ierr_code

      ! setup permutation vector P 
        allocate(P(nz), stat=ierr_code)
        print *, "allocation status code for P =", ierr_code


       !$acc enter data copyin(P, pBuffer)      ! P, pBuffer enter in

      ! Creates an identity map
        !$acc host_data use_device(P)
          status = cusparseCreateIdentityPermutation(cusparse_H, nz, P)
        !$acc end host_data
        print *, "CreateIdentityPermutation status =", status 

      ! sorts COO format by row
        !$acc host_data use_device(indx, jndx, P, pBuffer)
          status = cusparseXcoosortByRow(cusparse_H, m, n, nz, indx, jndx,  P, pBuffer)
        !$acc end host_data
        print *, "XcoosortByRow status =", status 

      ! gather sorted COO Vals = set "csrValA"
        !$acc host_data use_device(rval, csrValA, P)
          status = cusparseDgthr(cusparse_H, nz, rval, csrValA, P, CUSPARSE_INDEX_BASE_ZERO)
        !$acc end host_data

        !$acc exit data delete(P, pBuffer)      ! P, Pbuffer deleted

      ! index base is changed from 0 to 1 for subsequent Fortran cusolver routines on device
        !$acc kernels
          do i = 1, nz
            indx(i) = indx(i) + 1
            jndx(i) = jndx(i) + 1
          end do
        !$acc end kernels

      ! COO to CSR format = set "csrRowPtrA"
        !$acc host_data use_device(indx, csrRowPtrA)
         status = cusparseXcoo2csr(cusparse_H, indx, nz, m, csrRowPtrA, CUSPARSE_INDEX_BASE_ONE)
        !$acc end host_data

    !$acc end data

        deallocate(P, pBuffer)

      ! copy jndx to csrColIndA = set "csrColIndA"
        csrColIndA = jndx

      if (field == "complex") then
        call matrix_printout (field, nz, indx, jndx, cval1=cval)
      elseif (field == "real") then
        call matrix_printout (field, nz, indx, jndx, rval1=csrValA)
      end if

      ! check CSR data
      call csr_printout

    
  end subroutine coo2csr_convert

end module mm_coo_cuda_csr_convert
